@@ -242,7 +242,7 @@ inline static void BTC_memory_account_mark(NewGC *gc, mpage *page, void *ptr)
   if(page->size_class) {
     if(page->size_class > 1) {
       /* big page */
-      struct objhead *info = (struct objhead *)(NUM(page->addr) + PREFIX_SIZE);
+      objhead *info = BIG_PAGE_TO_OBJHEAD(page);
       
       if(info->btc_mark == gc->old_btc_mark) {
         info->btc_mark = gc->new_btc_mark;
@@ -251,17 +251,17 @@ inline static void BTC_memory_account_mark(NewGC *gc, mpage *page, void *ptr)
       }
     } else {
       /* medium page */
-      struct objhead *info = MED_OBJHEAD(ptr, page->size);
+      objhead *info = MED_OBJHEAD(ptr, page->size);
 
       if(info->btc_mark == gc->old_btc_mark) {
         info->btc_mark = gc->new_btc_mark;
         account_memory(gc, gc->current_mark_owner, info->size);
-        ptr = PTR(NUM(info) + WORD_SIZE);
+        ptr = OBJHEAD_TO_OBJPTR(info);
         push_ptr(ptr);
       }
     }
   } else {
-    struct objhead *info = (struct objhead *)((char*)ptr - WORD_SIZE);
+    objhead *info = OBJPTR_TO_OBJHEAD(ptr);
 
     if(info->btc_mark == gc->old_btc_mark) {
       info->btc_mark = gc->new_btc_mark;
@@ -302,7 +302,7 @@ int BTC_thread_mark(void *p)
 {
   NewGC *gc = GC_get_GC();
   if (gc->doing_memory_accounting) {
-    return ((struct objhead *)(NUM(p) - WORD_SIZE))->size;
+    return OBJPTR_TO_OBJHEAD(p)->size;
   }
   return gc->mark_table[btc_redirect_thread](p);
 }
@@ -314,7 +314,7 @@ int BTC_custodian_mark(void *p)
     if(custodian_to_owner_set(gc, p) == gc->current_mark_owner)
       return gc->mark_table[btc_redirect_custodian](p);
     else
-      return ((struct objhead *)(NUM(p) - WORD_SIZE))->size;
+      return OBJPTR_TO_OBJHEAD(p)->size;
   }
   return gc->mark_table[btc_redirect_custodian](p);
 }
@@ -323,7 +323,7 @@ int BTC_cust_box_mark(void *p)
 {
   NewGC *gc = GC_get_GC();
   if (gc->doing_memory_accounting) {
-    return ((struct objhead *)(NUM(p) - WORD_SIZE))->size;
+    return OBJPTR_TO_OBJHEAD(p)->size;
   }
   return gc->mark_table[btc_redirect_cust_box](p);
 }
@@ -338,21 +338,30 @@ inline static void mark_normal_obj(NewGC *gc, int type, void *ptr)
                            ignore them outright. In the case of custodians, we do need
                            to do the check; those differences are handled by replacing
                            the mark procedure in mark_table. */
-                        gc->mark_table[*(unsigned short*)ptr](ptr);
+                        unsigned short tag = *(unsigned short*)ptr;
+                        ASSERT_TAG(tag);
+                        if((unsigned long)gc->mark_table[tag] < PAGE_TYPES) {
+                          /* atomic */
+                        } else {
+                          GC_ASSERT(mark_table[tag]);
+                          gc->mark_table[tag](ptr);
+                        }
                         break;
                       }
     case PAGE_ATOMIC: break;
     case PAGE_ARRAY: { 
-                       struct objhead *info = (struct objhead *)((char*)ptr - WORD_SIZE);
-                       void **temp = ptr, **end = temp + (info->size - 1);
+                       objhead *info = OBJPTR_TO_OBJHEAD(ptr);
+                       void **temp = ptr;
+                       void **end  = PPTR(info) + info->size;
 
                        while(temp < end) gcMARK(*(temp++));
                        break;
                      };
     case PAGE_TARRAY: {
-                        struct objhead *info = (struct objhead *)((char*)ptr - WORD_SIZE);
+                        objhead *info = OBJPTR_TO_OBJHEAD(ptr);
                         unsigned short tag = *(unsigned short*)ptr;
-                        void **temp = ptr, **end = PPTR(info) + (info->size - INSET_WORDS);
+                        void **temp = ptr;
+                        void **end = PPTR(info) + (info->size - INSET_WORDS);
 
                         while(temp < end) temp += gc->mark_table[tag](temp);
                         break;
@@ -363,7 +372,7 @@ inline static void mark_normal_obj(NewGC *gc, int type, void *ptr)
 
 inline static void mark_acc_big_page(NewGC *gc, mpage *page)
 {
-  void **start = PPTR(NUM(page->addr) + PREFIX_SIZE + WORD_SIZE);
+  void **start = PPTR(BIG_PAGE_TO_OBJECT(page));
   void **end = PPTR(NUM(page->addr) + page->size);
 
   switch(page->page_type) {
@@ -401,14 +410,20 @@ static void propagate_accounting_marks(NewGC *gc)
   PageMap pagemap = gc->page_maps;
   while(pop_ptr(&p) && !gc->kill_propagation_loop) {
     page = pagemap_find_page(pagemap, p);
+#ifdef MZ_USE_PLACES
+    if (!page) {
+      page = pagemap_find_page(MASTERGC->page_maps, p);
+    }
+#endif
+    GC_ASSERT(page);
     set_backtrace_source(p, page->page_type);
     GCDEBUG((DEBUGOUTF, "btc_account: popped off page %p:%p, ptr %p\n", page, page->addr, p));
     if(page->size_class) {
       if (page->size_class > 1)
         mark_acc_big_page(gc, page);
       else {
-        struct objhead *info = MED_OBJHEAD(p, page->size);
-        p = PTR(NUM(info) + WORD_SIZE);
+        objhead *info = MED_OBJHEAD(p, page->size);
+        p = OBJHEAD_TO_OBJPTR(info);
         mark_normal_obj(gc, info->type, p);
       }
     } else
@@ -700,7 +715,7 @@ static inline void BTC_clean_up(NewGC *gc) {
   clean_up_account_hooks(gc);
 }
 
-static inline void BTC_set_btc_mark(NewGC *gc, void* x) {
-  ((struct objhead *)(x))->btc_mark = gc->old_btc_mark;
+static inline void BTC_set_btc_mark(NewGC *gc, objhead* info) {
+  info->btc_mark = gc->old_btc_mark;
 }
 #endif
