@@ -155,6 +155,7 @@
 # if defined(i386)
 #  define SCHEME_PLATFORM_LIBRARY_SUBPATH "i386-linux"
 #  define REGISTER_POOR_MACHINE
+#  define ASM_DBLPREC_CONTROL_87
 # endif
 # if defined(powerpc)
 #  define SCHEME_PLATFORM_LIBRARY_SUBPATH "ppc-linux"
@@ -178,6 +179,7 @@
 # if defined(__x86_64__)
 #  define SCHEME_PLATFORM_LIBRARY_SUBPATH "x86_64-linux"
 #  define REGISTER_POOR_MACHINE
+#  define ASM_DBLPREC_CONTROL_87
 # endif
 # ifndef SCHEME_PLATFORM_LIBRARY_SUBPATH
 #  define SCHEME_PLATFORM_LIBRARY_SUBPATH "unknown-linux"
@@ -214,6 +216,7 @@
 #if defined(__x86_64__)
 # define MZ_USE_JIT_X86_64
 # define MZ_JIT_USE_MPROTECT
+# define MZ_USE_DWARF_LIBUNWIND
 #endif
 #if defined(powerpc)
 # define MZ_USE_JIT_PPC
@@ -642,13 +645,21 @@
 # ifdef __POWERPC__
 #  define SCHEME_PLATFORM_LIBRARY_SUBPATH "ppc-darwin"
 # else
-#  define SCHEME_PLATFORM_LIBRARY_SUBPATH "i386-darwin"
+#  ifdef __x86_64__
+#   define SCHEME_PLATFORM_LIBRARY_SUBPATH "x86_64-darwin"
+#  else
+#   define SCHEME_PLATFORM_LIBRARY_SUBPATH "i386-darwin"
+#  endif
 # endif
 #else
 # ifdef __POWERPC__
 #  define SCHEME_PLATFORM_LIBRARY_SUBPATH "ppc-macosx"
 # else
-#  define SCHEME_PLATFORM_LIBRARY_SUBPATH "i386-macosx"
+#  ifdef __x86_64__
+#   define SCHEME_PLATFORM_LIBRARY_SUBPATH "x86_64-macosx"
+#  else
+#   define SCHEME_PLATFORM_LIBRARY_SUBPATH "i386-macosx"
+#  endif
 # endif
 #endif
 
@@ -690,11 +701,15 @@
 #  define OS_X 1
 #endif
 
-#ifdef __POWERPC__
+#if defined(__POWERPC__)
 # define MZ_USE_JIT_PPC
+#elif defined(__x86_64__)
+# define MZ_USE_JIT_X86_64
 #else
 # define MZ_USE_JIT_I386
+# define ASM_DBLPREC_CONTROL_87
 #endif
+# define MZ_JIT_USE_MPROTECT
 
 # define FLAGS_ALREADY_SET
 
@@ -1170,6 +1185,10 @@
     converts (a == a) to TRUE, even if `a' is floating-point. Used
     only when USE_[SCO_]IEEE_FP_PREDS is not defined. */
 
+ /* ASM_DBLPREC_CONTROL_87 uses inline assembly to set Intel '387
+    floating-point operations to double-precision instead of
+    extended-precision arithmetic. */
+
  /* IGNORE_BY_BORLAND_CONTROL_87 turns off floating-point error for
     Intel '387 with Borlad-style _control87. DONT_IGNORE_PIPE_SIGNAL
     can be on or off. */
