@@ -1,6 +1,6 @@
 /*
   MzScheme
-  Copyright (c) 2004-2008 PLT Scheme Inc.
+  Copyright (c) 2004-2009 PLT Scheme Inc.
   Copyright (c) 1995-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -27,6 +27,11 @@
 
 /* globals */
 Scheme_Object scheme_null[1];
+Scheme_Object *scheme_cons_proc;
+Scheme_Object *scheme_mcons_proc;
+Scheme_Object *scheme_list_proc;
+Scheme_Object *scheme_list_star_proc;
+Scheme_Object *scheme_box_proc;
 
 /* locals */
 static Scheme_Object *pair_p_prim (int argc, Scheme_Object *argv[]);
@@ -83,14 +88,18 @@ static Scheme_Object *set_box (int argc, Scheme_Object *argv[]);
 
 static Scheme_Object *make_hash(int argc, Scheme_Object *argv[]);
 static Scheme_Object *make_hasheq(int argc, Scheme_Object *argv[]);
+static Scheme_Object *make_hasheqv(int argc, Scheme_Object *argv[]);
 static Scheme_Object *make_weak_hash(int argc, Scheme_Object *argv[]);
 static Scheme_Object *make_weak_hasheq(int argc, Scheme_Object *argv[]);
+static Scheme_Object *make_weak_hasheqv(int argc, Scheme_Object *argv[]);
 static Scheme_Object *make_immutable_hash(int argc, Scheme_Object *argv[]);
 static Scheme_Object *make_immutable_hasheq(int argc, Scheme_Object *argv[]);
+static Scheme_Object *make_immutable_hasheqv(int argc, Scheme_Object *argv[]);
 static Scheme_Object *hash_table_count(int argc, Scheme_Object *argv[]);
 static Scheme_Object *hash_table_copy(int argc, Scheme_Object *argv[]);
 static Scheme_Object *hash_p(int argc, Scheme_Object *argv[]);
 static Scheme_Object *hash_eq_p(int argc, Scheme_Object *argv[]);
+static Scheme_Object *hash_eqv_p(int argc, Scheme_Object *argv[]);
 static Scheme_Object *hash_weak_p(int argc, Scheme_Object *argv[]);
 static Scheme_Object *hash_table_put_bang(int argc, Scheme_Object *argv[]);
 static Scheme_Object *hash_table_put(int argc, Scheme_Object *argv[]);
@@ -106,6 +115,7 @@ static Scheme_Object *hash_table_iterate_key(int argc, Scheme_Object *argv[]);
 static Scheme_Object *eq_hash_code(int argc, Scheme_Object *argv[]);
 static Scheme_Object *equal_hash_code(int argc, Scheme_Object *argv[]);
 static Scheme_Object *equal_hash2_code(int argc, Scheme_Object *argv[]);
+static Scheme_Object *eqv_hash_code(int argc, Scheme_Object *argv[]);
 
 static Scheme_Object *make_weak_box(int argc, Scheme_Object *argv[]);
 static Scheme_Object *weak_box_value(int argc, Scheme_Object *argv[]);
@@ -122,8 +132,18 @@ static Scheme_Object *placeholder_get(int argc, Scheme_Object *argv[]);
 static Scheme_Object *placeholder_p(int argc, Scheme_Object *argv[]);
 static Scheme_Object *make_hash_placeholder(int argc, Scheme_Object *argv[]);
 static Scheme_Object *make_hasheq_placeholder(int argc, Scheme_Object *argv[]);
+static Scheme_Object *make_hasheqv_placeholder(int argc, Scheme_Object *argv[]);
 static Scheme_Object *table_placeholder_p(int argc, Scheme_Object *argv[]);
 
+static Scheme_Object *unsafe_car (int argc, Scheme_Object *argv[]);
+static Scheme_Object *unsafe_cdr (int argc, Scheme_Object *argv[]);
+static Scheme_Object *unsafe_mcar (int argc, Scheme_Object *argv[]);
+static Scheme_Object *unsafe_mcdr (int argc, Scheme_Object *argv[]);
+static Scheme_Object *unsafe_set_mcar (int argc, Scheme_Object *argv[]);
+static Scheme_Object *unsafe_set_mcdr (int argc, Scheme_Object *argv[]);
+static Scheme_Object *unsafe_unbox (int argc, Scheme_Object *argv[]);
+static Scheme_Object *unsafe_set_box (int argc, Scheme_Object *argv[]);
+
 #define BOX "box"
 #define BOXP "box?"
 #define UNBOX "unbox"
@@ -148,7 +168,9 @@ scheme_init_list (Scheme_Env *env)
   SCHEME_PRIM_PROC_FLAGS(p) |= SCHEME_PRIM_IS_UNARY_INLINED;
   scheme_add_global_constant ("mpair?", p, env);
 
-  p = scheme_make_noncm_prim(cons_prim, "cons", 2, 2);
+  REGISTER_SO(scheme_cons_proc);
+  p = scheme_make_immed_prim(cons_prim, "cons", 2, 2);
+  scheme_cons_proc = p;
   SCHEME_PRIM_PROC_FLAGS(p) |= SCHEME_PRIM_IS_BINARY_INLINED;
   scheme_add_global_constant ("cons", p, env);
 
@@ -160,7 +182,9 @@ scheme_init_list (Scheme_Env *env)
   SCHEME_PRIM_PROC_FLAGS(p) |= SCHEME_PRIM_IS_UNARY_INLINED;
   scheme_add_global_constant ("cdr", p, env);
 
-  p = scheme_make_noncm_prim(mcons_prim, "mcons", 2, 2);
+  REGISTER_SO(scheme_mcons_proc);
+  p = scheme_make_immed_prim(mcons_prim, "mcons", 2, 2);
+  scheme_mcons_proc = p;
   SCHEME_PRIM_PROC_FLAGS(p) |= SCHEME_PRIM_IS_BINARY_INLINED;
   scheme_add_global_constant ("mcons", p, env);
 
@@ -189,16 +213,23 @@ scheme_init_list (Scheme_Env *env)
 						     "list?",
 						     1, 1),
 			      env);
-  scheme_add_global_constant ("list",
-			      scheme_make_immed_prim(list_prim,
-						     "list",
-						     0, -1),
-			      env);
-  scheme_add_global_constant ("list*",
-			      scheme_make_immed_prim(list_star_prim,
-						     "list*",
-						     1, -1),
-			      env);
+
+  REGISTER_SO(scheme_list_proc);
+  p = scheme_make_immed_prim(list_prim, "list", 0, -1);
+  scheme_list_proc = p;
+  SCHEME_PRIM_PROC_FLAGS(p) |= (SCHEME_PRIM_IS_UNARY_INLINED
+                                | SCHEME_PRIM_IS_BINARY_INLINED
+                                | SCHEME_PRIM_IS_NARY_INLINED);
+  scheme_add_global_constant ("list", p, env);
+
+  REGISTER_SO(scheme_list_star_proc);
+  p = scheme_make_immed_prim(list_star_prim, "list*", 1, -1);
+  scheme_list_star_proc = p;
+  SCHEME_PRIM_PROC_FLAGS(p) |= (SCHEME_PRIM_IS_UNARY_INLINED
+                                | SCHEME_PRIM_IS_BINARY_INLINED
+                                | SCHEME_PRIM_IS_NARY_INLINED);
+  scheme_add_global_constant ("list*", p, env);
+
   scheme_add_global_constant("immutable?",
 			     scheme_make_folding_prim(immutablep,
 						      "immutable?",
@@ -398,11 +429,12 @@ scheme_init_list (Scheme_Env *env)
                                                        1, 1, 1),
 			      env);
 
-  scheme_add_global_constant(BOX,
-			     scheme_make_immed_prim(box,
-						    BOX,
-						    1, 1),
-			     env);
+  REGISTER_SO(scheme_box_proc);
+  p = scheme_make_immed_prim(box, BOX, 1, 1);
+  scheme_box_proc = p;
+  SCHEME_PRIM_PROC_FLAGS(p) |= SCHEME_PRIM_IS_UNARY_INLINED;  
+  scheme_add_global_constant(BOX, p, env);
+
   scheme_add_global_constant("box-immutable",
 			     scheme_make_immed_prim(immutable_box,
 						    "box-immutable",
@@ -433,6 +465,11 @@ scheme_init_list (Scheme_Env *env)
 						    "make-hasheq",
 						    0, 0),
 			     env);
+  scheme_add_global_constant("make-hasheqv",
+			     scheme_make_immed_prim(make_hasheqv,
+						    "make-hasheqv",
+						    0, 0),
+			     env);
   scheme_add_global_constant("make-weak-hash",
 			     scheme_make_immed_prim(make_weak_hash,
 						    "make-weak-hash",
@@ -443,6 +480,11 @@ scheme_init_list (Scheme_Env *env)
 						    "make-weak-hasheq",
 						    0, 0),
 			     env);
+  scheme_add_global_constant("make-weak-hasheqv",
+			     scheme_make_immed_prim(make_weak_hasheqv,
+						    "make-weak-hasheqv",
+						    0, 0),
+			     env);
   scheme_add_global_constant("make-immutable-hash",
 			     scheme_make_immed_prim(make_immutable_hash,
 						    "make-immutable-hash",
@@ -453,6 +495,11 @@ scheme_init_list (Scheme_Env *env)
 						    "make-immutable-hasheq",
 						    1, 1),
 			     env);
+  scheme_add_global_constant("make-immutable-hasheqv",
+			     scheme_make_immed_prim(make_immutable_hasheqv,
+						    "make-immutable-hasheqv",
+						    1, 1),
+			     env);
   scheme_add_global_constant("hash?",
 			     scheme_make_folding_prim(hash_p,
 						      "hash?",
@@ -463,6 +510,11 @@ scheme_init_list (Scheme_Env *env)
 						      "hash-eq?",
 						      1, 1, 1),
 			     env);
+  scheme_add_global_constant("hash-eqv?",
+			     scheme_make_folding_prim(hash_eqv_p,
+						      "hash-eqv?",
+						      1, 1, 1),
+			     env);
   scheme_add_global_constant("hash-weak?",
 			     scheme_make_folding_prim(hash_weak_p,
 						      "hash-weak?",
@@ -540,6 +592,11 @@ scheme_init_list (Scheme_Env *env)
 						    "eq-hash-code",
 						    1, 1),
 			     env);
+  scheme_add_global_constant("eqv-hash-code",
+			     scheme_make_immed_prim(eqv_hash_code,
+						    "eqv-hash-code",
+						    1, 1),
+			     env);
   scheme_add_global_constant("equal-hash-code",
 			     scheme_make_noncm_prim(equal_hash_code,
 						    "equal-hash-code",
@@ -618,6 +675,11 @@ scheme_init_list (Scheme_Env *env)
 						      "make-hasheq-placeholder",
 						      1, 1),
 			     env);
+  scheme_add_global_constant("make-hasheqv-placeholder",
+			     scheme_make_prim_w_arity(make_hasheqv_placeholder,
+						      "make-hasheqv-placeholder",
+						      1, 1),
+			     env);
   scheme_add_global_constant("hash-placeholder?",
 			     scheme_make_folding_prim(table_placeholder_p,
 						      "hash-placeholder?",
@@ -631,6 +693,46 @@ scheme_init_list (Scheme_Env *env)
   equal_symbol = scheme_intern_symbol("equal");
 }
 
+void
+scheme_init_unsafe_list (Scheme_Env *env)
+{
+  Scheme_Object *p;
+  
+  scheme_null->type = scheme_null_type;
+
+  p = scheme_make_folding_prim(unsafe_car, "unsafe-car", 1, 1, 1);
+  SCHEME_PRIM_PROC_FLAGS(p) |= SCHEME_PRIM_IS_UNARY_INLINED;
+  scheme_add_global_constant ("unsafe-car", p, env);
+
+  p = scheme_make_folding_prim(unsafe_cdr, "unsafe-cdr", 1, 1, 1);
+  SCHEME_PRIM_PROC_FLAGS(p) |= SCHEME_PRIM_IS_UNARY_INLINED;
+  scheme_add_global_constant ("unsafe-cdr", p, env);
+
+  p = scheme_make_immed_prim(unsafe_mcar, "unsafe-mcar", 1, 1);
+  SCHEME_PRIM_PROC_FLAGS(p) |= SCHEME_PRIM_IS_UNARY_INLINED;
+  scheme_add_global_constant ("unsafe-mcar", p, env);
+
+  p = scheme_make_immed_prim(unsafe_mcdr, "unsafe-mcdr", 1, 1);
+  SCHEME_PRIM_PROC_FLAGS(p) |= SCHEME_PRIM_IS_UNARY_INLINED;
+  scheme_add_global_constant ("unsafe-mcdr", p, env);
+
+  p = scheme_make_immed_prim(unsafe_set_mcar, "unsafe-set-mcar!", 2, 2);
+  SCHEME_PRIM_PROC_FLAGS(p) |= SCHEME_PRIM_IS_BINARY_INLINED;
+  scheme_add_global_constant ("unsafe-set-mcar!", p, env);
+
+  p = scheme_make_immed_prim(unsafe_set_mcdr, "unsafe-set-mcdr!", 2, 2);
+  SCHEME_PRIM_PROC_FLAGS(p) |= SCHEME_PRIM_IS_BINARY_INLINED;
+  scheme_add_global_constant ("unsafe-set-mcdr!", p, env);
+  
+  p = scheme_make_immed_prim(unsafe_unbox, "unsafe-unbox", 1, 1);
+  SCHEME_PRIM_PROC_FLAGS(p) |= SCHEME_PRIM_IS_UNARY_INLINED;  
+  scheme_add_global_constant("unsafe-unbox", p, env);
+
+  p = scheme_make_immed_prim(unsafe_set_box, "unsafe-set-box!", 2, 2);
+  SCHEME_PRIM_PROC_FLAGS(p) |= SCHEME_PRIM_IS_BINARY_INLINED;
+  scheme_add_global_constant("unsafe-set-box!", p, env);
+}
+
 Scheme_Object *scheme_make_pair(Scheme_Object *car, Scheme_Object *cdr)
 {
 #ifdef MZ_PRECISE_GC
@@ -647,16 +749,12 @@ Scheme_Object *scheme_make_pair(Scheme_Object *car, Scheme_Object *cdr)
 
 Scheme_Object *scheme_make_mutable_pair(Scheme_Object *car, Scheme_Object *cdr)
 {
-#ifdef MZ_PRECISE_GC
-  return GC_malloc_mutable_pair(car, cdr);
-#else
   Scheme_Object *cons;
   cons = scheme_alloc_object();
   cons->type = scheme_mutable_pair_type;
   SCHEME_CAR(cons) = car;
   SCHEME_CDR(cons) = cdr;
   return cons;
-#endif
 }
 
 Scheme_Object *scheme_make_raw_pair(Scheme_Object *car, Scheme_Object *cdr)
@@ -674,11 +772,7 @@ Scheme_Object *scheme_make_raw_pair(Scheme_Object *car, Scheme_Object *cdr)
   return cons;
 }
 
-#ifdef MZ_PRECISE_GC
-# define cons(car, cdr) GC_malloc_pair(car, cdr)
-#else
 # define cons(car, cdr) scheme_make_pair(car, cdr)
-#endif
 
 Scheme_Object *scheme_build_list(int size, Scheme_Object **argv)
 {
@@ -974,7 +1068,7 @@ list_p_prim (int argc, Scheme_Object *argv[])
 
 #define LIST_BODY(INIT)                          \
   int i;                                         \
-  Scheme_Object *l;                              \
+  GC_CAN_IGNORE Scheme_Object *l;                \
   INIT;                                          \
   for (i = argc ; i--; ) {                       \
     l = cons(argv[i], l);                        \
@@ -1417,6 +1511,19 @@ static void make_hash_indices_for_equal(void *v, long *_stk_h1, long *_stk_h2)
     *_stk_h2 = scheme_equal_hash_key2((Scheme_Object *)v);
 }
 
+static int compare_eqv(void *v1, void *v2)
+{
+  return !scheme_eqv((Scheme_Object *)v1, (Scheme_Object *)v2);
+}
+
+static void make_hash_indices_for_eqv(void *v, long *_stk_h1, long *_stk_h2)
+{
+  if (_stk_h1)
+    *_stk_h1 = scheme_eqv_hash_key((Scheme_Object *)v);
+  if (_stk_h2)
+    *_stk_h2 = scheme_eqv_hash_key2((Scheme_Object *)v);
+}
+
 Scheme_Bucket_Table *scheme_make_weak_equal_table(void)
 {
   Scheme_Object *sema;
@@ -1432,6 +1539,21 @@ Scheme_Bucket_Table *scheme_make_weak_equal_table(void)
   return t;
 }
 
+Scheme_Bucket_Table *scheme_make_weak_eqv_table(void)
+{
+  Scheme_Object *sema;
+  Scheme_Bucket_Table *t;
+  
+  t = scheme_make_bucket_table(20, SCHEME_hash_weak_ptr);
+  
+  sema = scheme_make_sema(1);
+  t->mutex = sema;
+  t->compare = compare_eqv;
+  t->make_hash_indices = make_hash_indices_for_eqv;
+
+  return t;
+}
+
 static Scheme_Object *make_hash(int argc, Scheme_Object *argv[])
 {
   return (Scheme_Object *)scheme_make_hash_table_equal();
@@ -1442,6 +1564,11 @@ static Scheme_Object *make_hasheq(int argc, Scheme_Object *argv[])
   return (Scheme_Object *)scheme_make_hash_table(SCHEME_hash_ptr);
 }
 
+static Scheme_Object *make_hasheqv(int argc, Scheme_Object *argv[])
+{
+  return (Scheme_Object *)scheme_make_hash_table_eqv();
+}
+
 static Scheme_Object *make_weak_hash(int argc, Scheme_Object *argv[])
 {
   return (Scheme_Object *)scheme_make_weak_equal_table();
@@ -1452,7 +1579,12 @@ static Scheme_Object *make_weak_hasheq(int argc, Scheme_Object *argv[])
   return (Scheme_Object *)scheme_make_bucket_table(20, SCHEME_hash_weak_ptr);
 }
 
-static Scheme_Object *make_immutable_table(const char *who, int eql, int argc, Scheme_Object *argv[])
+static Scheme_Object *make_weak_hasheqv(int argc, Scheme_Object *argv[])
+{
+  return (Scheme_Object *)scheme_make_weak_eqv_table();
+}
+
+static Scheme_Object *make_immutable_table(const char *who, int kind, int argc, Scheme_Object *argv[])
 {
   Scheme_Object *l = argv[0], *a;
   Scheme_Hash_Tree *ht;
@@ -1468,7 +1600,7 @@ static Scheme_Object *make_immutable_table(const char *who, int eql, int argc, S
   if (!SCHEME_NULLP(l))
     scheme_wrong_type("make-immutable-hash", "list of pairs", 0, argc, argv);
 
-  ht = scheme_make_hash_tree(eql);
+  ht = scheme_make_hash_tree(kind);
 
   for (l = argv[0]; SCHEME_PAIRP(l); l = SCHEME_CDR(l)) {
     a = SCHEME_CAR(l);
@@ -1488,6 +1620,11 @@ static Scheme_Object *make_immutable_hasheq(int argc, Scheme_Object *argv[])
   return make_immutable_table("make-immutable-hasheq", 0, argc, argv);
 }
 
+static Scheme_Object *make_immutable_hasheqv(int argc, Scheme_Object *argv[])
+{
+  return make_immutable_table("make-immutable-hasheqv", 2, argc, argv);
+}
+
 Scheme_Hash_Table *scheme_make_hash_table_equal()
 {
   Scheme_Hash_Table *t;
@@ -1503,6 +1640,21 @@ Scheme_Hash_Table *scheme_make_hash_table_equal()
   return t;
 }
 
+Scheme_Hash_Table *scheme_make_hash_table_eqv()
+{
+  Scheme_Hash_Table *t;
+  Scheme_Object *sema;
+
+  t = scheme_make_hash_table(SCHEME_hash_ptr);
+
+  sema = scheme_make_sema(1);
+  t->mutex = sema;
+  t->compare = compare_eqv;
+  t->make_hash_indices = make_hash_indices_for_eqv;
+
+  return t;
+}
+
 static Scheme_Object *hash_table_count(int argc, Scheme_Object *argv[])
 {
   if (SCHEME_HASHTP(argv[0])) {
@@ -1595,13 +1747,15 @@ static Scheme_Object *hash_eq_p(int argc, Scheme_Object *argv[])
   Scheme_Object *o = argv[0];
 
   if (SCHEME_HASHTP(o)) {
-    if (((Scheme_Hash_Table *)o)->compare != compare_equal)
+    if ((((Scheme_Hash_Table *)o)->compare != compare_equal)
+        && (((Scheme_Hash_Table *)o)->compare != compare_eqv))
       return scheme_true;
   } else if (SCHEME_HASHTRP(o)) {
-    if (!(SCHEME_HASHTR_FLAGS((Scheme_Hash_Tree *)o) & 0x1))
+    if (!(SCHEME_HASHTR_FLAGS((Scheme_Hash_Tree *)o) & 0x3))
       return scheme_true;
   } else if (SCHEME_BUCKTP(o)) {
-    if (((Scheme_Bucket_Table *)o)->compare != compare_equal)
+    if ((((Scheme_Bucket_Table *)o)->compare != compare_equal)
+        && (((Scheme_Bucket_Table *)o)->compare != compare_eqv))
       return scheme_true;
   } else {
     scheme_wrong_type("hash-eq?", "hash", 0, argc, argv);
@@ -1610,6 +1764,26 @@ static Scheme_Object *hash_eq_p(int argc, Scheme_Object *argv[])
   return scheme_false;
 }
 
+static Scheme_Object *hash_eqv_p(int argc, Scheme_Object *argv[])
+{
+  Scheme_Object *o = argv[0];
+
+  if (SCHEME_HASHTP(o)) {
+    if (((Scheme_Hash_Table *)o)->compare == compare_eqv)
+      return scheme_true;
+  } else if (SCHEME_HASHTRP(o)) {
+    if (SCHEME_HASHTR_FLAGS((Scheme_Hash_Tree *)o) & 0x2)
+      return scheme_true;
+  } else if (SCHEME_BUCKTP(o)) {
+    if (((Scheme_Bucket_Table *)o)->compare == compare_eqv)
+      return scheme_true;
+  } else {
+    scheme_wrong_type("hash-eqv?", "hash", 0, argc, argv);
+  }
+  
+  return scheme_false;
+}
+
 static Scheme_Object *hash_weak_p(int argc, Scheme_Object *argv[])
 {
   Scheme_Object *o = argv[0];
@@ -1629,11 +1803,21 @@ int scheme_is_hash_table_equal(Scheme_Object *o)
   return (((Scheme_Hash_Table *)o)->compare == compare_equal);
 }
 
+int scheme_is_hash_table_eqv(Scheme_Object *o)
+{
+  return (((Scheme_Hash_Table *)o)->compare == compare_eqv);
+}
+
 int scheme_is_hash_tree_equal(Scheme_Object *o)
 {
   return SCHEME_HASHTR_FLAGS((Scheme_Hash_Tree *)o) & 0x1;
 }
 
+int scheme_is_hash_tree_eqv(Scheme_Object *o)
+{
+  return SCHEME_HASHTR_FLAGS((Scheme_Hash_Tree *)o) & 0x2;
+}
+
 static Scheme_Object *hash_table_put_bang(int argc, Scheme_Object *argv[])
 {
   Scheme_Object *v = argv[0];
@@ -2060,6 +2244,18 @@ static Scheme_Object *equal_hash2_code(int argc, Scheme_Object *argv[])
   return scheme_make_integer(v);
 }
 
+static Scheme_Object *eqv_hash_code(int argc, Scheme_Object *argv[])
+{
+  long v;
+
+  if (SCHEME_INTP(argv[0]))
+    return argv[0];
+
+  v = scheme_eqv_hash_key(argv[0]);
+
+  return scheme_make_integer(v);
+}
+
 Scheme_Object *scheme_make_weak_box(Scheme_Object *v)
 {
 #ifdef MZ_PRECISE_GC
@@ -2145,7 +2341,7 @@ static Scheme_Object *placeholder_p(int c, Scheme_Object *p[])
           : scheme_false);
 }
 
-static Scheme_Object *do_make_hash_placeholder(const char *who, int eq, int argc, Scheme_Object *argv[])
+static Scheme_Object *do_make_hash_placeholder(const char *who, int kind, int argc, Scheme_Object *argv[])
 {
   Scheme_Object *l, *a, *ph;
 
@@ -2162,14 +2358,14 @@ static Scheme_Object *do_make_hash_placeholder(const char *who, int eq, int argc
   ph = scheme_alloc_object();
   ph->type = scheme_table_placeholder_type;
   SCHEME_IPTR_VAL(ph) = argv[0];
-  SCHEME_PINT_VAL(ph) = eq;
+  SCHEME_PINT_VAL(ph) = kind;
 
   return ph;
 }
 
 static Scheme_Object *make_hash_placeholder(int argc, Scheme_Object *argv[])
 {
-  return do_make_hash_placeholder("make-hash-placeholder", 0, argc, argv);
+  return do_make_hash_placeholder("make-hash-placeholder", 1, argc, argv);
 }
 
 static Scheme_Object *make_hasheq_placeholder(int argc, Scheme_Object *argv[])
@@ -2177,6 +2373,11 @@ static Scheme_Object *make_hasheq_placeholder(int argc, Scheme_Object *argv[])
   return do_make_hash_placeholder("make-hash-placeholder", 0, argc, argv);
 }
 
+static Scheme_Object *make_hasheqv_placeholder(int argc, Scheme_Object *argv[])
+{
+  return do_make_hash_placeholder("make-hasheqv-placeholder", 2, argc, argv);
+}
+
 static Scheme_Object *table_placeholder_p(int c, Scheme_Object *p[])
 {
   return (SAME_TYPE(SCHEME_TYPE(p[0]), scheme_table_placeholder_type)
@@ -2382,3 +2583,50 @@ void scheme_init_ephemerons(void)
 }
 
 #endif
+
+/************************************************************/
+/*                        unsafe                            */
+/************************************************************/
+
+static Scheme_Object *unsafe_car (int argc, Scheme_Object *argv[])
+{
+  return SCHEME_CAR(argv[0]);
+}
+
+static Scheme_Object *unsafe_cdr (int argc, Scheme_Object *argv[])
+{
+  return SCHEME_CDR(argv[0]);
+}
+
+static Scheme_Object *unsafe_mcar (int argc, Scheme_Object *argv[])
+{
+  return SCHEME_CAR(argv[0]);
+}
+
+static Scheme_Object *unsafe_mcdr (int argc, Scheme_Object *argv[])
+{
+  return SCHEME_CDR(argv[0]);
+}
+
+static Scheme_Object *unsafe_set_mcar (int argc, Scheme_Object *argv[])
+{
+  SCHEME_CAR(argv[0]) = argv[1];
+  return scheme_void;
+}
+
+static Scheme_Object *unsafe_set_mcdr (int argc, Scheme_Object *argv[])
+{
+  SCHEME_CDR(argv[0]) = argv[1];
+  return scheme_void;
+}
+
+static Scheme_Object *unsafe_unbox (int argc, Scheme_Object *argv[])
+{
+  return SCHEME_BOX_VAL(argv[0]);
+}
+
+static Scheme_Object *unsafe_set_box (int argc, Scheme_Object *argv[])
+{
+  SCHEME_BOX_VAL(argv[0]) = argv[1];
+  return scheme_void;
+}
