@@ -355,8 +355,10 @@ cont_proc {
   MARK_cjs(&c->cjs);
   MARK_stack_state(&c->ss);
   gcMARK(c->barrier_prompt);
-  gcMARK(c->runstack_start);
-  gcMARK(c->runstack_saved);
+  if (!GC_merely_accounting()) {
+    gcMARK(c->runstack_start);
+    gcMARK(c->runstack_saved);
+  }
 
   gcMARK(c->prompt_id);
   gcMARK(c->prompt_buf);
@@ -615,12 +617,16 @@ thread_val {
   gcMARK(pr->init_config);
   gcMARK(pr->init_break_cell);
 
-  {
+  if (!pr->runstack_owner
+      || !GC_merely_accounting()
+      || (*pr->runstack_owner == pr)) {
     Scheme_Object **rs = pr->runstack_start;
     gcFIXUP_TYPED_NOW(Scheme_Object **, pr->runstack_start);
-    pr->runstack = pr->runstack_start + (pr->runstack - rs);
+    if (pr->runstack != pr->runstack_start + (pr->runstack - rs))
+      pr->runstack = pr->runstack_start + (pr->runstack - rs);
+
+    gcMARK(pr->runstack_saved);
   }
-  gcMARK(pr->runstack_saved);
   gcMARK(pr->runstack_owner);
   gcMARK(pr->runstack_swapped);
   pr->spare_runstack = NULL; /* just in case */
@@ -641,6 +647,9 @@ thread_val {
   
   gcMARK(pr->blocker);
   gcMARK(pr->overflow);
+
+  gcMARK(pr->return_marks_to);
+  gcMARK(pr->returned_marks);
   
   gcMARK(pr->current_local_env);
   gcMARK(pr->current_local_mark);
@@ -653,6 +662,7 @@ thread_val {
   gcMARK(pr->current_mt);
 
   gcMARK(pr->constant_folding);
+  gcMARK(pr->reading_delayed);
   
   gcMARK(pr->overflow_reply);
 
@@ -680,6 +690,7 @@ thread_val {
   gcMARK(pr->private_kill_next);
   
   gcMARK(pr->user_tls);
+  gcMARK(pr->gmp_tls_data);
   
   gcMARK(pr->mr_hop);
   gcMARK(pr->mref);
@@ -735,7 +746,8 @@ prompt_val {
  mark: 
   Scheme_Prompt *pr = (Scheme_Prompt *)p;
   gcMARK(pr->boundary_overflow_id);
-  gcMARK(pr->runstack_boundary_start);
+  if (!GC_merely_accounting())
+    gcMARK(pr->runstack_boundary_start);
   gcMARK(pr->tag);
   gcMARK(pr->id);
  size:
@@ -820,6 +832,8 @@ namespace_val {
   gcMARK(e->insp);
 
   gcMARK(e->rename_set);
+  gcMARK(e->temp_marked_names);
+  gcMARK(e->post_ex_rename_set);
 
   gcMARK(e->syntax);
   gcMARK(e->exp_env);
@@ -834,6 +848,9 @@ namespace_val {
   gcMARK(e->tt_require_names);
   gcMARK(e->dt_require_names);
   gcMARK(e->other_require_names);
+  gcMARK(e->did_starts);
+  gcMARK(e->available_next[0]);
+  gcMARK(e->available_next[1]);
 
   gcMARK(e->toplevel);
   gcMARK(e->modchain);
@@ -866,7 +883,8 @@ resolve_prefix_val {
   Resolve_Prefix *rp = (Resolve_Prefix *)p;
   gcMARK(rp->toplevels);
   gcMARK(rp->stxes);
-  gcMARK(rp->delay_info);
+  gcMARK(rp->delay_info_rpair);
+  gcMARK(rp->uses_unsafe);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Resolve_Prefix));
@@ -877,6 +895,7 @@ comp_prefix_val {
   Comp_Prefix *cp = (Comp_Prefix *)p;
   gcMARK(cp->toplevels);
   gcMARK(cp->stxes);
+  gcMARK(cp->uses_unsafe);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Comp_Prefix));
@@ -933,6 +952,8 @@ module_val {
   gcMARK(m->provide_protects);
   gcMARK(m->indirect_provides);
 
+  gcMARK(m->indirect_syntax_provides);
+
   gcMARK(m->et_provide_protects);
   gcMARK(m->et_indirect_provides);
 
@@ -943,6 +964,8 @@ module_val {
 
   gcMARK(m->insp);
 
+  gcMARK(m->lang_info);
+
   gcMARK(m->hints);
   gcMARK(m->ii_src);
 
@@ -970,6 +993,7 @@ module_phase_exports_val {
   gcMARK(m->provide_src_names);
   gcMARK(m->provide_nominal_srcs);
   gcMARK(m->provide_src_phases);
+  gcMARK(m->provide_insps);
 
   gcMARK(m->kernel_exclusion);
   gcMARK(m->kernel_exclusion2);
@@ -1063,7 +1087,7 @@ mark_logger {
 mark_log_reader {
  mark:
   Scheme_Log_Reader *lr = (Scheme_Log_Reader *)p;
-  gcMARK(lr->ch);
+  gcMARK(lr->sema);
   gcMARK(lr->head);
   gcMARK(lr->tail);
  size:
@@ -1074,6 +1098,17 @@ END type;
 
 /**********************************************************************/
 
+START engine;
+
+engine_val {
+ mark:
+  Scheme_Engine *en = (Scheme_Engine *)p;
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_Engine));
+}
+
+END engine;
+
 START env;
 
 mark_comp_env {
@@ -1097,6 +1132,7 @@ mark_comp_env {
   gcMARK(e->data.const_names);
   gcMARK(e->data.const_vals);
   gcMARK(e->data.const_uids);
+  gcMARK(e->data.sealed);
   gcMARK(e->data.use);
   gcMARK(e->data.lifts);
 
@@ -1299,6 +1335,19 @@ END hash;
 
 /**********************************************************************/
 
+START places;
+
+place_val {
+ mark:
+  Scheme_Place *pr = (Scheme_Place *)p;
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_Place));
+}
+
+END places;
+
+/**********************************************************************/
+
 START portfun;
 
 mark_load_handler_data {
@@ -1391,14 +1440,6 @@ mark_input_file {
   gcBYTES_TO_WORDS(sizeof(Scheme_Input_File));
 }
 
-#if defined(WIN32_FD_HANDLES)
-mark_tcp_select_info {
- mark:
- size:
-  gcBYTES_TO_WORDS(sizeof(Tcp_Select_Info));
-}
-#endif
-
 mark_output_file {
  mark:
   Scheme_Output_File *o = (Scheme_Output_File *)p;
@@ -1700,6 +1741,7 @@ mark_syncing {
   gcMARK(w->wrapss);
   gcMARK(w->nackss);
   gcMARK(w->reposts);
+  gcMARK(w->accepts);
   gcMARK(w->disable_break);
 
  size:
@@ -2032,9 +2074,9 @@ mark_rename_table {
   gcMARK(rn->nomarshal_ht);
   gcMARK(rn->unmarshal_info);
   gcMARK(rn->shared_pes);
-  gcMARK(rn->plus_kernel_nominal_source);
   gcMARK(rn->set_identity);
   gcMARK(rn->marked_names);
+  gcMARK(rn->free_id_renames);
  size:
   gcBYTES_TO_WORDS(sizeof(Module_Renames));
 }
@@ -2087,11 +2129,27 @@ lex_rib {
  mark:
   Scheme_Lexical_Rib *rib = (Scheme_Lexical_Rib *)p;
   gcMARK(rib->rename);
+  gcMARK(rib->timestamp);
+  gcMARK(rib->sealed);
   gcMARK(rib->next);
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Lexical_Rib));
 }
 
+mark_free_id_info {
+ mark:
+  Scheme_Vector *vec = (Scheme_Vector *)p;
+  int i;
+  for (i = 8; i--; )
+    gcMARK(vec->els[i]);
+
+ size:
+  gcBYTES_TO_WORDS((sizeof(Scheme_Vector) 
+		    + ((8 - 1) * sizeof(Scheme_Object *))));
+}
+
+
+
 END stxobj;
 
 /**********************************************************************/
@@ -2135,8 +2193,10 @@ native_unclosed_proc {
   int i;
 
   gcMARK(d->u2.name);
-  for (i = d->retain_count; i--; ) {
-    gcMARK(d->retained[i]);
+  if (d->retained) {
+    for (i = SCHEME_INT_VAL(d->retained[0]); i--; ) {
+      gcMARK(d->retained[i]);
+    }
   }
   if (d->closure_size < 0) {
     gcMARK(d->u.arities);
