@@ -1,6 +1,6 @@
 /*
   MzScheme
-  Copyright (c) 2004-2008 PLT Scheme Inc.
+  Copyright (c) 2004-2009 PLT Scheme Inc.
   Copyright (c) 2000-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -155,6 +155,21 @@ typedef struct Scheme_UDP {
 
 #endif /* USE_TCP */
 
+#if defined(WINDOWS_PROCESSES) || defined(WINDOWS_FILE_HANDLES)
+# define DECL_OS_FDSET(n) fd_set n[1]
+# define INIT_DECL_OS_FDSET(n) /* empty */
+# define MZ_OS_FD_ZERO(p) FD_ZERO(p)
+# define MZ_OS_FD_SET(n, p) FD_SET(n, p)
+# define MZ_OS_FD_CLR(n, p) FD_CLR(n, p)
+#else
+# define DECL_OS_FDSET(n) DECL_FDSET(n, 1)
+# define INIT_DECL_OS_FDSET(n) INIT_DECL_FDSET(n, 1)
+# define MZ_OS_FD_ZERO(p) MZ_FD_ZERO(p)
+# define MZ_OS_FD_SET(n, p) MZ_FD_SET(n, p)
+# define MZ_OS_FD_CLR(n, p) MZ_FD_CLR(n, p)
+#endif
+#define MZ_OS_FD_ISSET(n, p) FD_ISSET(n, p)
+
 static Scheme_Object *tcp_connect(int argc, Scheme_Object *argv[]);
 static Scheme_Object *tcp_connect_break(int argc, Scheme_Object *argv[]);
 static Scheme_Object *tcp_listen(int argc, Scheme_Object *argv[]);
@@ -937,25 +952,25 @@ static int tcp_check_accept(Scheme_Object *_listener)
 #ifdef USE_SOCKETS_TCP
   tcp_t s, mx;
   listener_t *listener = (listener_t *)_listener;
-  DECL_FDSET(readfds, 1);
-  DECL_FDSET(exnfds, 1);
+  DECL_OS_FDSET(readfds);
+  DECL_OS_FDSET(exnfds);
   struct timeval time = {0, 0};
   int sr, i;
 
-  INIT_DECL_FDSET(readfds, 1);
-  INIT_DECL_FDSET(exnfds, 1);
+  INIT_DECL_OS_FDSET(readfds);
+  INIT_DECL_OS_FDSET(exnfds);
 
   if (LISTENER_WAS_CLOSED(listener))
     return 1;
 
-  MZ_FD_ZERO(readfds);
-  MZ_FD_ZERO(exnfds);
+  MZ_OS_FD_ZERO(readfds);
+  MZ_OS_FD_ZERO(exnfds);
 
   mx = 0;
   for (i = 0; i < listener->count; i++) {
     s = listener->s[i];
-    MZ_FD_SET(s, readfds);
-    MZ_FD_SET(s, exnfds);
+    MZ_OS_FD_SET(s, readfds);
+    MZ_OS_FD_SET(s, exnfds);
     if (s > mx)
       mx = s;
   }
@@ -967,8 +982,8 @@ static int tcp_check_accept(Scheme_Object *_listener)
   if (sr) {
     for (i = 0; i < listener->count; i++) {
       s = listener->s[i];
-      if (FD_ISSET(s, readfds)
-	  || FD_ISSET(s, exnfds))
+      if (MZ_OS_FD_ISSET(s, readfds)
+	  || MZ_OS_FD_ISSET(s, exnfds))
 	return i + 1;
     }
   }
@@ -1001,21 +1016,21 @@ static int tcp_check_connect(Scheme_Object *connector_p)
 {
 #ifdef USE_SOCKETS_TCP
   tcp_t s;
-  DECL_FDSET(writefds, 1);
-  DECL_FDSET(exnfds, 1);
+  DECL_OS_FDSET(writefds);
+  DECL_OS_FDSET(exnfds);
   struct timeval time = {0, 0};
   int sr;
 
-  INIT_DECL_FDSET(writefds, 1);
-  INIT_DECL_FDSET(exnfds, 1);
+  INIT_DECL_OS_FDSET(writefds);
+  INIT_DECL_OS_FDSET(exnfds);
 
   s = *(tcp_t *)connector_p;
 
-  MZ_FD_ZERO(writefds);
-  MZ_FD_ZERO(exnfds);
+  MZ_OS_FD_ZERO(writefds);
+  MZ_OS_FD_ZERO(exnfds);
 
-  MZ_FD_SET(s, writefds);
-  MZ_FD_SET(s, exnfds);
+  MZ_OS_FD_SET(s, writefds);
+  MZ_OS_FD_SET(s, exnfds);
     
   do {
     sr = select(s + 1, NULL, writefds, exnfds, &time);
@@ -1056,20 +1071,20 @@ static int tcp_check_write(Scheme_Object *port)
 #ifdef USE_SOCKETS_TCP
   {
     tcp_t s;
-    DECL_FDSET(writefds, 1);
-    DECL_FDSET(exnfds, 1);
+    DECL_OS_FDSET(writefds);
+    DECL_OS_FDSET(exnfds);
     struct timeval time = {0, 0};
     int sr;
     
-    INIT_DECL_FDSET(writefds, 1);
-    INIT_DECL_FDSET(exnfds, 1);
+    INIT_DECL_OS_FDSET(writefds);
+    INIT_DECL_OS_FDSET(exnfds);
     
     s = data->tcp;
     
-    MZ_FD_ZERO(writefds);
-    MZ_FD_SET(s, writefds);
-    MZ_FD_ZERO(exnfds);
-    MZ_FD_SET(s, exnfds);
+    MZ_OS_FD_ZERO(writefds);
+    MZ_OS_FD_SET(s, writefds);
+    MZ_OS_FD_ZERO(exnfds);
+    MZ_OS_FD_SET(s, exnfds);
     
     do {
       sr = select(s + 1, NULL, writefds, exnfds, &time);
@@ -1156,12 +1171,12 @@ static int tcp_byte_ready (Scheme_Input_Port *port)
   Scheme_Tcp *data;
 #ifdef USE_SOCKETS_TCP
   int sr;
-  DECL_FDSET(readfds, 1);
-  DECL_FDSET(exfds, 1);
+  DECL_OS_FDSET(readfds);
+  DECL_OS_FDSET(exfds);
   struct timeval time = {0, 0};
 
-  INIT_DECL_FDSET(readfds, 1);
-  INIT_DECL_FDSET(exfds, 1);
+  INIT_DECL_OS_FDSET(readfds);
+  INIT_DECL_OS_FDSET(exfds);
 #endif
 
   if (port->closed)
@@ -1175,10 +1190,10 @@ static int tcp_byte_ready (Scheme_Input_Port *port)
     return 1;
 
 #ifdef USE_SOCKETS_TCP
-  MZ_FD_ZERO(readfds);
-  MZ_FD_ZERO(exfds);
-  MZ_FD_SET(data->tcp, readfds);
-  MZ_FD_SET(data->tcp, exfds);
+  MZ_OS_FD_ZERO(readfds);
+  MZ_OS_FD_ZERO(exfds);
+  MZ_OS_FD_SET(data->tcp, readfds);
+  MZ_OS_FD_SET(data->tcp, exfds);
     
   do {
     sr = select(data->tcp + 1, readfds, NULL, exfds, &time);
@@ -1480,7 +1495,10 @@ static long tcp_write_string(Scheme_Output_Port *port,
 
   if (!len) {
     /* Flush */
-    return tcp_flush(port, rarely_block, enable_break);
+    tcp_flush(port, rarely_block, enable_break);
+    if (data->b.out_bufpos != data->b.out_bufmax)
+      return -1;
+    return 0;
   }
 
   if (rarely_block) {
@@ -1561,13 +1579,16 @@ tcp_out_buffer_mode(Scheme_Port *p, int mode)
 }
 
 static Scheme_Object *
-make_tcp_input_port(void *data, const char *name)
+make_tcp_input_port(void *data, const char *name, Scheme_Object *cust)
 {
   Scheme_Input_Port *ip;
+
+  if (cust)
+    scheme_set_next_port_custodian((Scheme_Custodian *)cust);
   
   ip = scheme_make_input_port(scheme_tcp_input_port_type,
 			      data,
-			      scheme_make_immutable_sized_utf8_string((char *)name, -1),
+			      scheme_intern_symbol(name),
 			      tcp_get_string,
 			      NULL,
 			      scheme_progress_evt_via_get,
@@ -1583,13 +1604,16 @@ make_tcp_input_port(void *data, const char *name)
 }
 
 static Scheme_Object *
-make_tcp_output_port(void *data, const char *name)
+make_tcp_output_port(void *data, const char *name, Scheme_Object *cust)
 {
   Scheme_Output_Port *op;
 
+  if (cust)
+    scheme_set_next_port_custodian((Scheme_Custodian *)cust);
+
   op = scheme_make_output_port(scheme_tcp_output_port_type,
 						  data,
-						  scheme_make_immutable_sized_utf8_string((char *)name, -1),
+						  scheme_intern_symbol(name),
 						  scheme_write_evt_via_write,
 						  tcp_write_string,
 						  (Scheme_Out_Ready_Fun)tcp_check_write,
@@ -1790,8 +1814,8 @@ static Scheme_Object *tcp_connect(int argc, Scheme_Object *argv[])
 
 	      tcp = make_tcp_port_data(s, 2);
 	      
-	      v[0] = make_tcp_input_port(tcp, address);
-	      v[1] = make_tcp_output_port(tcp, address);
+	      v[0] = make_tcp_input_port(tcp, address, NULL);
+	      v[1] = make_tcp_output_port(tcp, address, NULL);
 	      
 	      REGISTER_SOCKET(s);
 
@@ -2160,7 +2184,8 @@ tcp_accept_ready(int argc, Scheme_Object *argv[])
 }
 
 static Scheme_Object *
-tcp_accept(int argc, Scheme_Object *argv[])
+do_tcp_accept(int argc, Scheme_Object *argv[], Scheme_Object *cust, char **_fail_reason)
+/* If _fail_reason is not NULL, never raise an exception. */
 {
 #ifdef USE_TCP
   int was_closed = 0, errid, ready_pos;
@@ -2191,12 +2216,22 @@ tcp_accept(int argc, Scheme_Object *argv[])
     ready_pos = 0;
 
   if (was_closed) {
-    scheme_raise_exn(MZEXN_FAIL_NETWORK,
-		     "tcp-accept: listener is closed");
+    if (_fail_reason)
+      *_fail_reason = "tcp-accept-evt: listener is closed";
+    else
+      scheme_raise_exn(MZEXN_FAIL_NETWORK,
+                       "tcp-accept: listener is closed");
     return NULL;
   }
 
-  scheme_custodian_check_available(NULL, "tcp-accept", "network");
+  if (!_fail_reason)
+    scheme_custodian_check_available((Scheme_Custodian *)cust, "tcp-accept", "network");
+  else {
+    if (!scheme_custodian_is_available((Scheme_Custodian *)cust)) {
+      *_fail_reason = "tcp-accept-evt: custodian is shutdown";
+      return NULL;
+    }
+  }
   
 # ifdef USE_SOCKETS_TCP
   ls = ((listener_t *)listener)->s[ready_pos-1];
@@ -2220,8 +2255,8 @@ tcp_accept(int argc, Scheme_Object *argv[])
 
     tcp = make_tcp_port_data(s, 2);
 
-    v[0] = make_tcp_input_port(tcp, "[accepted]");
-    v[1] = make_tcp_output_port(tcp, "[accepted]");
+    v[0] = make_tcp_input_port(tcp, "tcp-accepted", cust);
+    v[1] = make_tcp_output_port(tcp, "tcp-accepted", cust);
 
     scheme_file_open_count++;
     REGISTER_SOCKET(s);
@@ -2231,8 +2266,11 @@ tcp_accept(int argc, Scheme_Object *argv[])
   errid = SOCK_ERRNO();
 # endif
   
-  scheme_raise_exn(MZEXN_FAIL_NETWORK,
-		   "tcp-accept: accept from listener failed (%E)", errid);
+  if (_fail_reason)
+    *_fail_reason = "tcp-accept-evt: accept from listener failed";
+  else
+    scheme_raise_exn(MZEXN_FAIL_NETWORK,
+                     "tcp-accept: accept from listener failed (%E)", errid);
 #else
   scheme_wrong_type("tcp-accept", "tcp-listener", 0, argc, argv);
 #endif
@@ -2240,6 +2278,12 @@ tcp_accept(int argc, Scheme_Object *argv[])
   return NULL;
 }
 
+static Scheme_Object *
+tcp_accept(int argc, Scheme_Object *argv[])
+{
+  return do_tcp_accept(argc, argv, NULL, NULL);
+}
+
 static Scheme_Object *
 tcp_accept_break(int argc, Scheme_Object *argv[])
 {
@@ -2445,35 +2489,54 @@ static Scheme_Object *tcp_port_p(int argc, Scheme_Object *argv[])
 
 static Scheme_Object *tcp_accept_evt(int argc, Scheme_Object *argv[])
 {
-  Scheme_Object *r;
+  Scheme_Object *r, *custodian;
 
   if (!SAME_TYPE(SCHEME_TYPE(argv[0]), scheme_listener_type))
     scheme_wrong_type("tcp-accept-evt", "tcp-listener", 0, argc, argv);
 
-  r = scheme_alloc_small_object();
+  custodian = scheme_get_param(scheme_current_config(), MZCONFIG_CUSTODIAN);
+
+  scheme_custodian_check_available((Scheme_Custodian *)custodian, "tcp-accept", "network");
+  
+  r = scheme_alloc_object();
   r->type = scheme_tcp_accept_evt_type;
-  SCHEME_PTR_VAL(r) = argv[0];
+  SCHEME_PTR1_VAL(r) = argv[0];
+  SCHEME_PTR2_VAL(r) = custodian;
 
   return r;
 }
 
+static Scheme_Object *accept_failed(void *msg, int argc, Scheme_Object **argv)
+{
+  scheme_raise_exn(MZEXN_FAIL_NETWORK, msg ? (const char *)msg : "accept failed");
+  return NULL;
+} 
+
 static int tcp_check_accept_evt(Scheme_Object *ae, Scheme_Schedule_Info *sinfo)
 {
-  if (tcp_check_accept(SCHEME_PTR_VAL(ae))) {
+  if (tcp_check_accept(SCHEME_PTR1_VAL(ae))) {
     Scheme_Object *a[2];
-    a[0] = SCHEME_PTR_VAL(ae);
-    tcp_accept(1, a);
-    a[0] = scheme_current_thread->ku.multiple.array[0];
-    a[1] = scheme_current_thread->ku.multiple.array[1];
-    scheme_set_sync_target(sinfo, scheme_build_list(2, a), NULL, NULL, 0, 0);
-    return 1;
+    char *fail_reason = NULL;
+    a[0] = SCHEME_PTR1_VAL(ae);
+    if (do_tcp_accept(1, a, SCHEME_PTR2_VAL(ae), &fail_reason)) {
+      a[0] = scheme_current_thread->ku.multiple.array[0];
+      a[1] = scheme_current_thread->ku.multiple.array[1];
+      scheme_set_sync_target(sinfo, scheme_build_list(2, a), NULL, NULL, 0, 0, NULL);
+      return 1;
+    } else {
+      /* error on accept */
+      scheme_set_sync_target(sinfo, scheme_always_ready_evt, 
+                             scheme_make_closed_prim(accept_failed, fail_reason), 
+                             NULL, 0, 0, NULL);
+      return 1;
+    }
   } else
     return 0;
 }
 
 static void tcp_accept_evt_needs_wakeup(Scheme_Object *ae, void *fds)
 {
-  tcp_accept_needs_wakeup(SCHEME_PTR_VAL(ae), fds);
+  tcp_accept_needs_wakeup(SCHEME_PTR1_VAL(ae), fds);
 }
 
 int scheme_get_port_socket(Scheme_Object *p, long *_s)
@@ -2518,9 +2581,9 @@ void scheme_socket_to_ports(long s, const char *name, int takeover,
 
   tcp = make_tcp_port_data(s, takeover ? 2 : 3);
 
-  v = make_tcp_input_port(tcp, name);
+  v = make_tcp_input_port(tcp, name, NULL);
   *_inp = v;
-  v = make_tcp_output_port(tcp, name);
+  v = make_tcp_output_port(tcp, name, NULL);
   *_outp = v;
   
   if (takeover) {
@@ -2880,18 +2943,18 @@ static int udp_check_send(Scheme_Object *_udp)
     return 1;
 
   {
-    DECL_FDSET(writefds, 1);
-    DECL_FDSET(exnfds, 1);
+    DECL_OS_FDSET(writefds);
+    DECL_OS_FDSET(exnfds);
     struct timeval time = {0, 0};
     int sr;
     
-    INIT_DECL_FDSET(writefds, 1);
-    INIT_DECL_FDSET(exnfds, 1);
+    INIT_DECL_OS_FDSET(writefds);
+    INIT_DECL_OS_FDSET(exnfds);
     
-    MZ_FD_ZERO(writefds);
-    MZ_FD_SET(udp->s, writefds);
-    MZ_FD_ZERO(exnfds);
-    MZ_FD_SET(udp->s, exnfds);
+    MZ_OS_FD_ZERO(writefds);
+    MZ_OS_FD_SET(udp->s, writefds);
+    MZ_OS_FD_ZERO(exnfds);
+    MZ_OS_FD_SET(udp->s, exnfds);
     
     do {
       sr = select(udp->s + 1, NULL, writefds, exnfds, &time);
@@ -3112,18 +3175,18 @@ static int udp_check_recv(Scheme_Object *_udp)
     return 1;
 
   {
-    DECL_FDSET(readfds, 1);
-    DECL_FDSET(exnfds, 1);
+    DECL_OS_FDSET(readfds);
+    DECL_OS_FDSET(exnfds);
     struct timeval time = {0, 0};
     int sr;
     
-    INIT_DECL_FDSET(readfds, 1);
-    INIT_DECL_FDSET(exnfds, 1);
+    INIT_DECL_OS_FDSET(readfds);
+    INIT_DECL_OS_FDSET(exnfds);
     
-    MZ_FD_ZERO(readfds);
-    MZ_FD_SET(udp->s, readfds);
-    MZ_FD_ZERO(exnfds);
-    MZ_FD_SET(udp->s, exnfds);
+    MZ_OS_FD_ZERO(readfds);
+    MZ_OS_FD_SET(udp->s, readfds);
+    MZ_OS_FD_ZERO(exnfds);
+    MZ_OS_FD_SET(udp->s, exnfds);
     
     do {
       sr = select(udp->s + 1, readfds, NULL, exnfds, &time);
@@ -3359,7 +3422,7 @@ static int udp_evt_check_ready(Scheme_Object *_uw, Scheme_Schedule_Info *sinfo)
       if (do_udp_recv("udp-receive!-evt", uw->udp, 
 		      uw->str, uw->offset, uw->offset + uw->len, 
 		      0, v)) {
-	scheme_set_sync_target(sinfo, scheme_build_list(3, v), NULL, NULL, 0, 0);
+	scheme_set_sync_target(sinfo, scheme_build_list(3, v), NULL, NULL, 0, 0, NULL);
 	return 1;
       } else
 	return 0;
@@ -3374,7 +3437,7 @@ static int udp_evt_check_ready(Scheme_Object *_uw, Scheme_Schedule_Info *sinfo)
 			 uw->dest_addr, uw->dest_addr_len,
 			 0);
       if (SCHEME_TRUEP(r)) {
-	scheme_set_sync_target(sinfo, scheme_void, NULL, NULL, 0, 0);
+	scheme_set_sync_target(sinfo, scheme_void, NULL, NULL, 0, 0, NULL);
 	return 1;
       } else
 	return 0;
