@@ -111,6 +111,7 @@ Scheme_Custodian_Reference *(*scheme_add_managed)(Scheme_Custodian *m, Scheme_Ob
 							 Scheme_Close_Custodian_Client *f, void *data,
 							 int strong);
 void (*scheme_custodian_check_available)(Scheme_Custodian *m, const char *who, const char *what);
+int (*scheme_custodian_is_available)(Scheme_Custodian *m);
 void (*scheme_remove_managed)(Scheme_Custodian_Reference *m, Scheme_Object *o);
 void (*scheme_close_managed)(Scheme_Custodian *m);
 void (*scheme_schedule_custodian_close)(Scheme_Custodian *c);
@@ -674,6 +675,7 @@ Scheme_Output_Port *(*scheme_make_output_port)(Scheme_Object *subtype, void *dat
 						      Scheme_Write_Special_Evt_Fun write_special_evt_fun,
 						      Scheme_Write_Special_Fun write_special_fun,
 						      int must_close);
+void (*scheme_set_next_port_custodian)(Scheme_Custodian *c);
 void (*scheme_set_port_location_fun)(Scheme_Port *port,
 					    Scheme_Location_Fun location_fun);
 void (*scheme_set_port_count_lines_fun)(Scheme_Port *port,
